import inspect
import logging
from typing import Mapping, Any

from _pytest.fixtures import FixtureRequest
import wrapt

import allure
from allure_commons.types import AttachmentType
from allure_commons.lifecycle import AllureLifecycle
from allure_commons.model2 import TestResult
from allure_commons import plugin_manager
from allure_commons.model2 import TestStepResult

from allure_pytest_bdd.pytest_bdd_listener import PytestBDDListener

from enhanced_reports.config import Parameter

logger = logging.getLogger(__name__)


logger.info("Loaded " + __file__)


def attach_text(name: str, value: str, **kwargs):
    allure.attach(bytes(value, "utf-8"), name, AttachmentType.TEXT)


def attach_image(name: str, path: str, **kwargs):
    allure.attach.file(
        path,
        name=name,
        attachment_type=AttachmentType.PNG,
    )


def attach_video(name: str, path: str, **kwargs):
    allure.attach.file(
        path,
        name=name,
        attachment_type=AttachmentType.WEBM,
    )


def update_test_results_for_scenario_outline():
    logger.debug("Entered " + inspect.currentframe().f_code.co_name)

    def _custom_write_test_case(self, uuid=None):
        """Allure has an open bug (https://github.com/allure-framework/allure-python/issues/636) which prevents the
        inclusion of tests with scenario outlines in allure report. There is no fix available yet, so we manually
        remove the params which are equals to '_pytest_bdd_example'. This param if included in test results, causes
        errors in report generation due to duplicate keys in the json"""
        test_result = self._pop_item(uuid=uuid, item_type=TestResult)
        if test_result:
            if test_result.parameters:
                adj_parameters = []
                for param in test_result.parameters:
                    if param.name != "_pytest_bdd_example":
                        # do not include parameters with "_pytest_bdd_example"
                        adj_parameters.append(param)
                test_result.parameters = adj_parameters

            for index in range(0, len(test_result.steps)):
                if ':\n|' in test_result.steps[index].name:
                    # Remove the data table from the step name
                    step_desc = test_result.steps[index].name.split(':')
                    test_result.steps[index].name = step_desc[0]

            plugin_manager.hook.report_result(result=test_result)

    AllureLifecycle.write_test_case = _custom_write_test_case


def wrapper_for_unexecuted_steps():
    """When a bdd step fails, test execution is stopped hence next steps are not executed,
    allure report doesn't include the steps that were not executed due to a failed step before them
    To overcome this issue we are intercepting the PytestBDDListener._scenario_finalizer method to add the
    non executed steps to test results"""
    logger.debug("Entered " + inspect.currentframe().f_code.co_name)

    @wrapt.patch_function_wrapper(PytestBDDListener, "_scenario_finalizer")
    def wrap_scenario_finalizer(wrapped, instance, args, kwargs):
        # here, wrapped is the original perform method in PytestBDDListener
        # instance is `self` (it is not the case for classmethods though),
        # args and kwargs are a tuple and a dict respectively.

        wrapped(*args, **kwargs)  # note it is already bound to the instance

        test_result = instance.lifecycle._get_item(
            uuid=instance.lifecycle._last_item_uuid(item_type=TestResult),
            item_type=TestResult,
        )
        if len(args[0].steps) > len(test_result.steps):
            # if there are more steps in scenario than in test result, then add the remaining steps to test result
            for i in range(len(test_result.steps), len(args[0].steps)):
                test_result.steps.append(
                    TestStepResult(
                        name=f"{args[0].steps[i].keyword} {args[0].steps[i].name}",
                        status="skipped",
                    )
                )


def perform_session_setup(
    request: FixtureRequest, report_options: Mapping[Parameter, Any]
):
    update_test_results_for_scenario_outline()
    wrapper_for_unexecuted_steps()


def perform_session_cleanup(
    request: FixtureRequest, report_options: Mapping[Parameter, Any]
):
    pass


def perform_function_setup(
    request: FixtureRequest, report_options: Mapping[Parameter, Any]
):
    pass


def perform_function_cleanup(
    request: FixtureRequest, report_options: Mapping[Parameter, Any]
):
    pass



from selenium.webdriver.remote.webdriver import WebDriver
from datetime import datetime
import logging
import inspect


logger = logging.getLogger(__name__)
logger.info("Loaded " + __file__)


def get_js_logs(driver: WebDriver):
    logger.debug("Entered " + inspect.currentframe().f_code.co_name)
    logs = _capture_output(driver)
    return _format_outputs(logs)


def _capture_output(driver: WebDriver):
    return driver.get_log("browser")


def _format_outputs(logs: list) -> str:
    output = ""
    for item in logs:
        i_time = datetime.fromtimestamp(
            item.get("timestamp", 0) / 1000
        ).strftime("%Y-%m-%d %H:%M:%S")
        i_source = item.get("source")
        i_level = item.get("level")
        i_message = item.get("message")
        output += "{} {} {} {}\n".format(i_time, i_level, i_source, i_message)

    return output



import os
from enum import Enum
from typing import Mapping, Any, Union

from _pytest.config import argparsing
from _pytest.fixtures import FixtureRequest

import logging


logger = logging.getLogger(__name__)
logger.info("Loaded " + __file__)


class EnhancedReportOperationFrequency(Enum):
    ALWAYS = "always"
    EACH_UI_OPERATION = "each_ui_operation"
    END_OF_EACH_TEST = "end_of_each_test"
    FAILED_TEST_ONLY = "failed_test_only"
    NEVER = "never"


class Parameter(Enum):
    JS_LOG_FREQUENCY = "browser_console_log_capture"

    SS_FREQUENCY = "screenshot_capture"
    SS_RESIZE_PERCENT = "screenshot_resize_percent"
    SS_HEIGHT = "screenshot_height"
    SS_WIDTH = "screenshot_width"
    SS_HIGHLIGHT_ELEMENT = "highlighted_screenshot"
    SS_KEEP_FILES = "keep_screenshots"
    SS_DIR = "screenshot_dir"

    VIDEO_ENABLED = "video_recording"
    VIDEO_KEEP_FILES = "keep_videos"
    VIDEO_DIR = "video_dir"
    VIDEO_RESIZE_PERCENT = "video_resize_percent"
    VIDEO_FRAME_RATE = "video_frame_rate"
    VIDEO_HEIGHT = "video_height"
    VIDEO_WIDTH = "video_width"


__default_action = "store"
__default_prefix = "report_"

__param_values: Mapping[Parameter, Any] = {}

__params: Mapping[Parameter, Mapping[str, Any]] = {
    Parameter.JS_LOG_FREQUENCY: {
        "default_value": "failed_test_only",
        "allowed_values": [
            "always",
            "each_ui_operation",
            "end_of_each_test",
            "failed_test_only",
            "never",
        ],
        "cast_to": EnhancedReportOperationFrequency,
        "doc": "Specifies when to capture info from the browser console log.",
    },
    Parameter.SS_FREQUENCY: {
        "default_value": "each_ui_operation",
        "allowed_values": [
            "always",
            "each_ui_operation",
            "end_of_each_test",
            "failed_test_only",
            "never",
        ],
        "cast_to": EnhancedReportOperationFrequency,
        "doc": "Specifies when to capture screenshots.",
    },
    Parameter.SS_RESIZE_PERCENT: {
        "default_value": 40,
        "doc": "A percentage by which the screenshot will be resized. This is ignored if screenshot height and width "
        "values are also provided. Valid values - 75, 60, 50, etc.",
        "cast_to": int,
    },
    Parameter.SS_HEIGHT: {
        "default_value": 0,
        "cast_to": int,
        "doc": "The expected height of the resized screenshot used in reports. Actual value could be slightly "
        "different as it needs to fit the aspect ratio.",
    },
    Parameter.SS_WIDTH: {
        "default_value": 0,
        "cast_to": int,
        "doc": "The expected width of the resized screenshot used in reports. Actual value could be slightly different "
        "as it needs to fit the aspect ratio.",
    },
    Parameter.SS_HIGHLIGHT_ELEMENT: {
        "default_value": False,
        "cast_to": bool,
        "doc": "If set to True, the element being interacted with will be highlighted before taking the screenshot.",
    },
    Parameter.SS_KEEP_FILES: {
        "default_value": False,
        "cast_to": bool,
        "doc": "If set to True, generated screenshot images will not be deleted after the test run.",
    },
    Parameter.SS_DIR: {
        "default_value": "reports/screenshots",
        "doc": "The path to the directory where screenshots will be stored.",
    },
    Parameter.VIDEO_ENABLED: {
        "default_value": False,
        "cast_to": bool,
        "doc": "If set to True, a video will be recorded for each test.",
    },
    Parameter.VIDEO_KEEP_FILES: {
        "default_value": False,
        "cast_to": bool,
        "doc": "If set to True, generated video files will not be deleted after the test run.",
    },
    Parameter.VIDEO_DIR: {
        "default_value": "reports/videos",
        "doc": "The path to the directory where video files will be stored.",
    },
    Parameter.VIDEO_RESIZE_PERCENT: {
        "default_value": 75,
        "cast_to": int,
        "doc": "A percentage by which the video frames will be resized. This is ignored if screenshot height and width "
        "values are also provided. Valid values - 75, 60, 50, etc.",
    },
    Parameter.VIDEO_FRAME_RATE: {
        "default_value": 30,
        "cast_to": int,
        "doc": "The expected number of frames per second while recording a video. This is applicable only when "
        "enough frames were recorded in one second, which is not guaranteed.",
    },
    Parameter.VIDEO_HEIGHT: {
        "default_value": 0,
        "cast_to": int,
        "doc": "Expected height of the video. Actual value could be different as it needs to fit the aspect ratio.",
    },
    Parameter.VIDEO_WIDTH: {
        "default_value": 0,
        "cast_to": int,
        "doc": "Expected width of the video. Actual value could be different as it needs to fit the aspect ratio.",
    },
}


def _get_value(request: FixtureRequest, parameter: Parameter):
    full_arg_name: str = f"{__default_prefix}{parameter.value}"
    val_from_env_var = os.getenv(
        full_arg_name.upper(), default=__params[parameter]["default_value"]
    )
    value = request.config.getoption(full_arg_name, default=val_from_env_var)

    # Cast the value to the specified type, if needed
    if "cast_to" in __params[parameter]:
        value = __params[parameter]["cast_to"](value)

    return value


def register_with(
    parser_or_group: Union[argparsing.Parser, argparsing.OptionGroup]
):
    """Adds the command line arguments to the parser"""
    for parameter, details in __params.items():

        # Generate the docstring for the argument
        docstring: str = details.get("doc", "No docstring for this argument")
        if "default_value" in details:
            def_value = details.get("default_value")

            # Surround with single quotes if it is a string
            if type(def_value) == str:
                def_value = f"'{def_value}'"

            docstring += f""" Default value - {def_value}."""
        if "allowed_values" in details:
            docstring += f""" Allowed values: '{"', '".join(details.get("allowed_values"))}'."""

        parser_or_group.addoption(
            f"--{__default_prefix}{parameter.value}",
            action=details.get("action", __default_action),
            default=details.get("default_value", None),
            help=docstring,
        )


def get_all_values(request: FixtureRequest):
    """Returns a dictionary with values for all the options"""
    global __param_values

    if __param_values:
        return __param_values

    logger.debug("Getting values for all the options")
    __param_values = {
        parameter: _get_value(request, parameter)
        for parameter in __params.keys()
    }
    return __param_values


def get_value(parameter: Parameter, request: FixtureRequest = None):
    """Gets the value for a specified argument.
    Looks for the argument in the following order:
    1. Command line argument
    2. Environment variable
    3. Default value, set by the plugin
    """
    return get_all_values(request)[parameter]



from typing import Dict, Any, Tuple

from datetime import datetime
import base64
from PIL import Image
from io import BytesIO

from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.remote.webelement import WebElement
from selenium.webdriver.support import expected_conditions

from . import common_utils
from .config import Parameter

import logging
import inspect

logger = logging.getLogger(__name__)
logger.info("Loaded " + __file__)


__desired_resolution: Tuple[int, int] = None
__resize_factor: float = None


@common_utils.fail_silently
def get_screenshot(
    screenshot_name: str, scenario_name: str, plugin_options, driver: WebDriver
) -> str:
    logger.debug("Entered " + inspect.currentframe().f_code.co_name)
    # selenium can't take screenshots if a browser alert/prompt is open. trying to do so would break the current test.
    # so, skipping screenshots in such a case
    if expected_conditions.alert_is_present()(driver):
        return ""
    return __get_resized_image(
        driver.get_screenshot_as_base64(),
        plugin_options,
        scenario_name,
        screenshot_name=screenshot_name,
    )


@common_utils.fail_silently
def get_highlighted_screenshot(
    element: WebElement,
    action_name: str,
    scenario_name: str,
    report_options,
    driver: WebDriver,
    color: str = "red",
    border_width: int = 5,
) -> str:
    logger.debug("Entered " + inspect.currentframe().f_code.co_name)

    def apply_style(s):
        driver.execute_script(
            "arguments[0].setAttribute('style', arguments[1]);", element, s
        )

    original_style = element.get_attribute("style")
    apply_style(
        "border: {0}px solid {1}; padding:{2}px".format(border_width, color, 5)
    )

    path: str = get_screenshot(
        action_name, scenario_name, report_options, driver
    )

    apply_style(original_style)

    return path


def __get_resized_image(
    image_bytes,
    report_options: Dict[Parameter, Any],
    scenario_name,
    screenshot_name="screenshot",
):
    global __desired_resolution, __resize_factor
    __desired_resolution = (
        __desired_resolution
        if __desired_resolution
        else (
            report_options[Parameter.SS_WIDTH],
            report_options[Parameter.SS_HEIGHT],
        )
    )
    __resize_factor = (
        __resize_factor
        if __resize_factor
        else report_options[Parameter.SS_RESIZE_PERCENT] / 100
    )

    screenshot_file_name: str = (
        common_utils.clean_filename(f" {str(datetime.now())}") + ".png"
    )
    logger.debug("Temp screenshot file name: " + screenshot_file_name)

    # open the image directly thru an in-memory buffer
    img = Image.open(BytesIO(base64.b64decode(image_bytes)))

    # Check if user wants to keep the screenshots, if yes then create directory and save original images
    if report_options[Parameter.SS_KEEP_FILES]:
        ss_dir: str = f"{report_options[Parameter.SS_DIR]}/{scenario_name}"
        common_utils.mkdir(ss_dir)
        img.save(f"{ss_dir}/{screenshot_file_name}")

    # if the user has not passed a specific resolution, create it from the resize factor
    desired_resolution = (
        common_utils.get_resized_resolution(
            img.width, img.height, __resize_factor
        )
        if __desired_resolution == (0, 0)
        else __desired_resolution
    )
    logger.debug("Desired resolution: " + str(desired_resolution))

    # resize image to the desired resolution. if more customizability is needed, consider the resize or reduce methods
    img.thumbnail(desired_resolution)
    # in tobytes() need to return the array before the join operation happens
    # return img.tobytes()

    path: str = (
        f"{report_options[Parameter.SS_DIR]}/"
        f"{common_utils.clean_filename(scenario_name)}_{screenshot_file_name}"
    )
    img.save(path)
    logger.debug("Resized screenshot file path: " + path)
    return path
